@@ -35,7 +35,7 @@
 #include <fcitx/keys.h>
 #include <fcitx/module.h>
 #include <fcitx/context.h>
-#include <fcitx/module/punc/punc.h>
+#include <fcitx/module/punc/fcitx-punc.h>
 #include <string>
 #include <libintl.h>
 
@@ -48,14 +48,10 @@
 #ifdef __cplusplus
 extern "C" {
 #endif
-    FCITX_EXPORT_API
-    FcitxIMClass ime = {
+    FCITX_DEFINE_PLUGIN(fcitx_sunpinyin, ime, FcitxIMClass) = {
         FcitxSunpinyinCreate,
         FcitxSunpinyinDestroy
     };
-
-    FCITX_EXPORT_API
-    int ABI_VERSION = FCITX_ABI_VERSION;
 #ifdef __cplusplus
 }
 #endif
@@ -88,13 +84,47 @@ static const char* fuzzyPairs[][2] = {
     {"k", "g"}
 };
 
-static const char *correctionPairs[][2] = {
+struct CorrectionPair {
+    const char* first;
+    const char* second;
+};
+
+static const CorrectionPair uenunPairs[] = {
+    {"uen", "un"},
+};
+
+static const CorrectionPair imgingPairs[] = {
+    {"img", "ing"},
+};
+
+static const CorrectionPair iouiuPairs[] = {
+    {"iou", "iu"},
+};
+
+static const CorrectionPair ueiuiPairs[] = {
+    {"uei", "ui"},
+};
+
+static const CorrectionPair gnngPairs[] = {
     {"ign", "ing"},
     {"ogn", "ong"},
-    {"uen", "un"},
-    {"img", "ing"},
-    {"iou", "iu"},
-    {"uei", "ui"}
+    {"agn", "ang"},
+    {"egn", "eng"},
+};
+
+struct CorrectionPairs {
+    const CorrectionPair* correctionPair;
+    size_t size;
+};
+
+#define _DEF_PAIR(NAME) {NAME, sizeof(NAME) / sizeof(NAME[0]) }
+
+static const CorrectionPairs correctionPairs[] = {
+    _DEF_PAIR(uenunPairs),
+    _DEF_PAIR(imgingPairs),
+    _DEF_PAIR(iouiuPairs),
+    _DEF_PAIR(ueiuiPairs),
+    _DEF_PAIR(gnngPairs),
 };
 
 /**
@@ -159,8 +189,34 @@ INPUT_RETURN_VALUE FcitxSunpinyinDoInput(void* arg, FcitxKeySym sym, unsigned in
     if (FcitxHotkeyIsHotKey(sym, state, FCITX_SPACE))
         return FcitxCandidateWordChooseByIndex(candList, 0);
 
+    if (FcitxHotkeyIsHotKey(sym, state, FCITX_ENTER)) {
+        if (FcitxInputStateGetRawInputBufferSize(input)) {
+            /* remove space from preedit */
+            char* p = sunpinyin->preedit;
+            char* pp = sunpinyin->preedit;
+
+            while (*p) {
+                uint32_t chr;
+
+                char* next = fcitx_utf8_get_char(p, &chr);
+                // space
+                if (chr != 0x20) {
+                    if (p != pp) {
+                        memmove(pp, p, next - p);
+                    }
+                    pp += next - p;
+                }
+                p = next;
+            }
+            *pp = 0;
+            FcitxInstanceCommitString(sunpinyin->owner, FcitxInstanceGetCurrentIC(sunpinyin->owner), sunpinyin->preedit);
+            return IRV_CLEAN;
+        } else {
+            return IRV_TO_PROCESS;
+        }
+    }
+
     if ((view->getIC()->isEmpty() || !sunpinyin->fs.bProcessPunc)
-        && !FcitxHotkeyIsHotKeyUAZ(sym, state)
         && !FcitxHotkeyIsHotKeyLAZ(sym, state)
         && !FcitxHotkeyIsHotKey(sym, state, FCITX_SEMICOLON)
         && !FcitxHotkeyIsHotKey(sym, state, FCITX_BACKSPACE)
@@ -219,6 +275,7 @@ INPUT_RETURN_VALUE FcitxSunpinyinGetCandWords(void* arg)
     FcitxInputState* input = FcitxInstanceGetInputState(instance);
     FcitxGlobalConfig* config = FcitxInstanceGetGlobalConfig(sunpinyin->owner);
     FcitxCandidateWordList* candList = FcitxInputStateGetCandidateList(input);
+    FcitxMessages* clientPreedit = FcitxInputStateGetClientPreedit(input);
     FcitxCandidateWordSetPageSize(candList, config->iMaxCandWord);
 
     CPreEditString ppd;
@@ -246,7 +303,7 @@ INPUT_RETURN_VALUE FcitxSunpinyinGetCandWords(void* arg)
 
     memset(sunpinyin->clientpreedit, 0, FCITX_SUNPINYIN_MAX(hzlen * UTF8_MAX_LENGTH + 1, MAX_USER_INPUT + 1));
     WCSTOMBS(sunpinyin->clientpreedit, sunpinyin->input_src, MAX_USER_INPUT);
-    FcitxMessagesAddMessageAtLast(FcitxInputStateGetClientPreedit(input), MSG_INPUT, "%s", sunpinyin->clientpreedit);
+    FcitxMessagesAddMessageAtLast(clientPreedit, MSG_INPUT, "%s", sunpinyin->clientpreedit);
     FcitxInputStateSetClientCursorPos(input, 0);
 
     memset(sunpinyin->preedit, 0, FCITX_SUNPINYIN_MAX(ppd.size() * UTF8_MAX_LENGTH + 1, MAX_USER_INPUT + 1));
@@ -332,7 +389,6 @@ void* FcitxSunpinyinCreate (FcitxInstance* instance)
         return NULL;
     }
 
-#if FCITX_CHECK_VERSION(4,2,1)
     /* portable detect here */
     if (getenv("FCITXDIR")) {
         char* path = fcitx_utils_get_fcitx_path_with_filename("libdir", "sunpinyin/data");
@@ -340,7 +396,6 @@ void* FcitxSunpinyinCreate (FcitxInstance* instance)
         ASimplifiedChinesePolicy::instance().setDataDir(spath);
         free(path);
     }
-#endif
 
     CSunpinyinSessionFactory& fac = CSunpinyinSessionFactory::getFactory();
 
@@ -386,11 +441,10 @@ void* FcitxSunpinyinCreate (FcitxInstance* instance)
                     ReloadConfigFcitxSunpinyin,
                     NULL,
                     1,
-                    "zh_CN"
-                   );
+                    "zh_CN");
 
-    AddFunction(addon, (void*) SunpinyinGetFullPinyin);
-    AddFunction(addon, (void*) SunpinyinAddWord);
+    FcitxModuleAddFunction(addon, SunpinyinGetFullPinyin);
+    FcitxModuleAddFunction(addon, SunpinyinAddWord);
 
     return sunpinyin;
 }
@@ -488,6 +542,8 @@ void ConfigSunpinyin(FcitxSunpinyin* sunpinyin)
         }
         sunpinyin->view->setCancelOnBackspace(1);
         sunpinyin->view->getIC()->setHistoryPower(fs->iMemoryStrength);
+        sunpinyin->view->getIC()->setMaxBest(fs->maxBest + 1);
+        sunpinyin->view->getIC()->setMaxTailCandidateNum(fs->maxTail);
     }
 
     string_pairs fuzzy, correction;
@@ -496,8 +552,12 @@ void ConfigSunpinyin(FcitxSunpinyin* sunpinyin)
             fuzzy.push_back(std::make_pair<std::string, std::string>(fuzzyPairs[i][0], fuzzyPairs[i][1]));
 
     for (i = 0; i < CORRECT_SIZE; i++)
-        if (fs->bAutoCorrecting[i])
-            correction.push_back(std::make_pair<std::string, std::string>(correctionPairs[i][0], correctionPairs[i][1]));
+        if (fs->bAutoCorrecting[i]) {
+            for (int j = 0; j < correctionPairs[i].size; j ++ ) {
+                correction.push_back(std::make_pair<std::string, std::string>(correctionPairs[i].correctionPair[j].first,
+                                                                              correctionPairs[i].correctionPair[j].second));
+            }
+        }
 
     if (fuzzy.size() != 0)
     {
@@ -576,15 +636,11 @@ void UpdatePunc(FcitxSunpinyin* sunpinyin)
         return;
     const char symbol[] = "!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~";
     string_pairs puncPairs;
-    for( int i = 0; i < sizeof(symbol)/sizeof(char) -1; i++) {
+    for (unsigned int i = 0;i < sizeof(symbol) / sizeof(char) - 1;i++) {
         int c = symbol[i];
         char s[2] = {symbol[i], '\0'};
-        char* p1 = NULL, *p2 = NULL;
-        FcitxModuleFunctionArg args;
-        args.args[0] = &c;
-        args.args[1] = &p1;
-        args.args[2] = &p2;
-        InvokeFunction(sunpinyin->owner, FCITX_PUNC, GETPUNC2, args);
+        char *p1 = NULL, *p2 = NULL;
+        FcitxPuncGetPunc2(sunpinyin->owner, &c, &p1, &p2);
         string_pair p;
         p.first = s;
         if (p1) {
@@ -621,9 +677,9 @@ void* SunpinyinAddWord(void* arg, FcitxModuleFunctionArg args)
         return NULL;
 
     /* no way to check real single character pronouce, but let it be here */
-    for (int i = 0; i < segments.size(); i ++) {
+    for (unsigned int i = 0;i < segments.size();i++) {
         const IPySegmentor::TSegment& segment = segments[i];
-        for (int j = 0; j < segment.m_syllables.size(); j ++) {
+        for (unsigned int j = 0;j < segment.m_syllables.size();j++) {
             TSyllable syl = segment.m_syllables[j];
             if (!syl.isFullSyllable())
                 return NULL;
